import unittest
import time
import threading
from concurrent.futures import ThreadPoolExecutor
from suggest.suggest import ThreadSafeMemoryCache


class TestThreadSafeMemoryCache(unittest.TestCase):

    def setUp(self):
        self.cache = ThreadSafeMemoryCache(max_size=5, ttl_seconds=1)

    def test_init(self):
        """Test initialization with different parameters."""
        cache1 = ThreadSafeMemoryCache()
        self.assertEqual(cache1.max_size, 1000)
        self.assertEqual(cache1.ttl, 3600)

        cache2 = ThreadSafeMemoryCache(max_size=10, ttl_seconds=60)
        self.assertEqual(cache2.max_size, 10)
        self.assertEqual(cache2.ttl, 60)

    def test_get_nonexistent(self):
        """Test get for non-existent key."""
        self.assertIsNone(self.cache.get("nonexistent"))

    def test_set_and_get(self):
        """Test setting and getting a value."""
        self.cache.set("key1", "value1")
        self.assertEqual(self.cache.get("key1"), "value1")

    def test_update_existing(self):
        """Test updating existing key."""
        self.cache.set("key1", "value1")
        self.cache.set("key1", "updated_value")
        self.assertEqual(self.cache.get("key1"), "updated_value")

    def test_expiration(self):
        """Test that items expire after TTL."""
        cache = ThreadSafeMemoryCache(ttl_seconds=0.1)
        cache.set("key1", "value1")
        self.assertEqual(cache.get("key1"), "value1")

        # Wait for expiration
        time.sleep(0.2)
        self.assertIsNone(cache.get("key1"))

    def test_lru_eviction(self):
        """Test LRU eviction when cache reaches max size."""
        # Fill cache to capacity
        for i in range(5):
            self.cache.set(f"key{i}", f"value{i}")

        # All items should be in cache
        for i in range(5):
            self.assertEqual(self.cache.get(f"key{i}"), f"value{i}")

        # Add one more item, should evict the oldest (key0)
        self.cache.set("new_key", "new_value")
        self.assertIsNone(self.cache.get("key0"))
        self.assertEqual(self.cache.get("new_key"), "new_value")

        # Accessing key1 should move it to the end of the LRU queue
        self.assertEqual(self.cache.get("key1"), "value1")

        # Adding another item should now evict key2, not key1
        self.cache.set("another_key", "another_value")
        self.assertIsNone(self.cache.get("key2"))
        self.assertEqual(self.cache.get("key1"), "value1")

    def test_clear(self):
        """Test clearing the cache."""
        self.cache.set("key1", "value1")
        self.cache.set("key2", "value2")

        self.cache.clear()

        self.assertIsNone(self.cache.get("key1"))
        self.assertIsNone(self.cache.get("key2"))

    def test_remove_expired(self):
        """Test removing expired items."""
        cache = ThreadSafeMemoryCache(ttl_seconds=0.1)
        cache.set("key1", "value1")
        cache.set("key2", "value2")

        # Wait for expiration
        time.sleep(0.2)

        # Add a fresh item
        cache.set("key3", "value3")

        # Remove expired items
        removed = cache.remove_expired()

        # Should have removed 2 items
        self.assertEqual(removed, 2)
        self.assertIsNone(cache.get("key1"))
        self.assertIsNone(cache.get("key2"))
        self.assertEqual(cache.get("key3"), "value3")

    def test_thread_safety(self):
        """Test thread safety with concurrent operations."""
        num_threads = 50
        iterations = 100

        def worker():
            for i in range(iterations):
                # Add unique keys for each thread/iteration
                key = f"thread_{threading.get_ident()}_{i}"
                self.cache.set(key, i)

                # Random read (may or may not exist)
                self.cache.get(f"thread_{i % num_threads}_{i % iterations}")

        # Run concurrent workers
        with ThreadPoolExecutor(max_workers=num_threads) as executor:
            futures = [executor.submit(worker) for _ in range(num_threads)]

            # Wait for all to complete
            for future in futures:
                future.result()

        # If we got here without exceptions, the test passed
        # We would expect lock issues to cause exceptions

    def test_key_generation(self):
        """Test key generation logic."""
        key1 = self.cache._get_key("arg1", "arg2", kwarg1="value1")
        key2 = self.cache._get_key("arg1", "arg2", kwarg1="value1")
        key3 = self.cache._get_key("arg1", "arg2", kwarg1="different")

        # Same args should produce same key
        self.assertEqual(key1, key2)

        # Different args should produce different keys
        self.assertNotEqual(key1, key3)

        # Order of kwargs shouldn't matter
        key4 = self.cache._get_key("arg1", kwarg2="value2", kwarg1="value1")
        key5 = self.cache._get_key("arg1", kwarg1="value1", kwarg2="value2")
        self.assertEqual(key4, key5)



import unittest
from unittest.mock import MagicMock
from geosupport import GeosupportError

from suggest import GeosupportSuggest


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Create mock Geosupport object
        mock_geosupport = MagicMock()

        # Configure the mock function with appropriate returns based on inputs
        def mock_geocode(**kwargs):
            house_number = kwargs.get("house_number")
            street = kwargs.get("street", "")
            borough_code = kwargs.get("borough_code")
            zip_code = kwargs.get("zip")

            # Case: Insufficient input (just '100')
            if house_number == "100" and not street:
                return None

            # Case: Gold st with borough code 1 (Manhattan)
            if (
                house_number
                and street
                and "GOLD" in street.upper()
                and borough_code == 1
            ):
                return {
                    "First Borough Name": "MANHATTAN",
                    "House Number - Display Format": house_number,
                    "First Street Name Normalized": "GOLD STREET",
                }

            # Case: Gold st with ZIP 10038 (Manhattan)
            if (
                house_number
                and street
                and "GOLD" in street.upper()
                and zip_code == "10038"
            ):
                return {
                    "First Borough Name": "MANHATTAN",
                    "House Number - Display Format": house_number,
                    "First Street Name Normalized": "GOLD STREET",
                }

            # Case: Similar names for 'Gol'
            if house_number and street and street.upper() == "GOL":
                # Important: Create GeosupportError correctly
                error = GeosupportError({})
                error.result = {
                    "Message": "SIMILAR NAMES",
                    "List of Street Names": ["GOLD STREET", "GOLD AVENUE"],
                }
                raise error

            # Default case for Gold st (multiple results by borough)
            if house_number and street and "GOLD" in street.upper():
                # For test_has_suggestions, ensure we return MANHATTAN for any GOLD search
                return {
                    "First Borough Name": "MANHATTAN",
                    "House Number - Display Format": house_number,
                    "First Street Name Normalized": "GOLD STREET",
                }

            return None

        # Set up the mock_function to be returned by __getitem__
        mock_function = MagicMock(side_effect=mock_geocode)
        mock_geosupport.__getitem__.return_value = mock_function

        cls.suggest = GeosupportSuggest(mock_geosupport)



import unittest
from suggest.suggest import AddressFormatter


class TestAddressFormatter(unittest.TestCase):

    def test_format_borough_normal(self):
        """Test formatting normal borough names."""
        self.assertEqual(AddressFormatter.format_borough("Manhattan"), "MANHATTAN")
        self.assertEqual(AddressFormatter.format_borough("queens"), "QUEENS")
        self.assertEqual(AddressFormatter.format_borough("Brooklyn"), "BROOKLYN")

    def test_format_borough_edge_cases(self):
        """Test formatting edge cases for borough names."""
        self.assertEqual(AddressFormatter.format_borough(None), "")
        self.assertEqual(AddressFormatter.format_borough(""), "")
        self.assertEqual(AddressFormatter.format_borough("   "), "   ")

    def test_format_bbl_dict(self):
        """Test formatting BBL from dictionary."""
        bbl_dict = {"BOROUGH BLOCK LOT (BBL)": "1002501001"}
        self.assertEqual(AddressFormatter.format_bbl(bbl_dict), "1002501001")

    def test_format_bbl_string(self):
        """Test formatting BBL from string."""
        self.assertEqual(AddressFormatter.format_bbl("1002501001"), "1002501001")

    def test_format_bbl_edge_cases(self):
        """Test formatting edge cases for BBL."""
        self.assertIsNone(AddressFormatter.format_bbl(None))
        self.assertEqual(AddressFormatter.format_bbl({}), None)
        self.assertEqual(AddressFormatter.format_bbl({"other_key": "value"}), None)

    def test_format_coordinates_normal(self):
        """Test formatting normal coordinates."""
        coords = AddressFormatter.format_coordinates(40.7128, -74.0060)
        self.assertEqual(coords["latitude"], 40.7128)
        self.assertEqual(coords["longitude"], -74.0060)

    def test_format_coordinates_strings(self):
        """Test formatting string coordinates."""
        coords = AddressFormatter.format_coordinates("40.7128", "-74.0060")
        self.assertEqual(coords["latitude"], 40.7128)
        self.assertEqual(coords["longitude"], -74.0060)

    def test_format_coordinates_edge_cases(self):
        """Test formatting edge cases for coordinates."""
        self.assertIsNone(AddressFormatter.format_coordinates(None, -74.0060))
        self.assertIsNone(AddressFormatter.format_coordinates(40.7128, None))
        self.assertIsNone(AddressFormatter.format_coordinates(None, None))

    def test_format_coordinates_invalid(self):
        """Test formatting invalid coordinates."""
        self.assertIsNone(AddressFormatter.format_coordinates("invalid", -74.0060))
        self.assertIsNone(AddressFormatter.format_coordinates(40.7128, "invalid"))
        self.assertIsNone(
            AddressFormatter.format_coordinates("invalid", "also_invalid")
        )



from tests.testcase import TestCase
from unittest.mock import MagicMock
from geosupport import GeosupportError
from suggest import GeosupportSuggest


class TestSuggestions(TestCase):

    def test_no_suggestions(self):
        result = self.suggest.suggestions("100")
        self.assertListEqual(result, [])

    def test_has_suggestions(self):
        result = self.suggest.suggestions("100 Gold st")
        self.assertTrue(len(result) >= 1)
        # Verify the mock returned expected data
        self.assertEqual(result[0]["First Borough Name"], "MANHATTAN")

    def test_suggestions_have_values(self):
        # Create a dedicated mock for this test
        mock_geosupport = MagicMock()
        mock_func = MagicMock()
        mock_geosupport.__getitem__.return_value = mock_func

        # Create a test-specific instance
        test_suggest = GeosupportSuggest(mock_geosupport)

        # Create a proper GeosupportError with the result attribute structure
        # that matches what's expected in your _geocode method
        mock_error = GeosupportError({})
        mock_error.result = {
            "Message": "SIMILAR NAMES",
            "List of Street Names": ["GOLD STREET", "GOLD AVENUE"],
        }

        # Only need 3 side effects - one for the error, two for the similar streets
        mock_func.side_effect = [
            mock_error,
            {
                "First Borough Name": "MANHATTAN",
                "House Number - Display Format": "100",
                "First Street Name Normalized": "GOLD STREET",
            },
            {
                "First Borough Name": "BROOKLYN",
                "House Number - Display Format": "100",
                "First Street Name Normalized": "GOLD AVENUE",
            },
        ]

        # Use a single borough call to simplify the test
        result = test_suggest.suggestions("100 Gol", borough_code=1)

        self.assertTrue(len(result) >= 1)
        for s in result:
            self.assertTrue(len(s["First Borough Name"]) >= 1)
            self.assertTrue(len(s["House Number - Display Format"]) >= 1)
            self.assertTrue(len(s["First Street Name Normalized"]) >= 1)

    def test_suggestions_with_zip_code(self):
        result = self.suggest.suggestions("100 Gold st, 10038")
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["First Borough Name"], "MANHATTAN")

    def test_suggestions_have_no_none(self):
        result = self.suggest.suggestions("100 Gold")
        self.assertTrue(None not in result)

    def test_borough_code_argument(self):
        result = self.suggest.suggestions("100 Gold st", borough_code=1)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]["First Borough Name"], "MANHATTAN")
        self.assertIsNotNone(result[0]["House Number - Display Format"])
        self.assertIsNotNone(result[0]["First Street Name Normalized"])



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()


setup(
    name="geosupport-suggest",
    version="0.1.0",
    url="https://github.com/ishiland/geosupport-suggest",
    description="Retrieve address suggestions from Geosupport",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ian Shiland",
    author_email="ishiland@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    keywords=[
        "NYC",
        "geocoder",
        "python-geosupport",
        "geosupport",
        "geosupport-suggest",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    python_requires=">=3.8",
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "geosupport-suggest=suggest.cli:main",
        ],
    },
    test_suite="tests",
    extras_require={
        "docs": [
            "sphinx>=4.0.0",
            "sphinx_rtd_theme>=1.0.0",
        ],
        "test": [
            "pytest",
            "pytest-cov",
        ],
        "dev": [
            "pytest",
            "pytest-cov",
            "black",
        ],
    },
)



from geosupport import GeosupportError
from nycparser import Parser
from typing import List, Dict, Any, Optional, Union, Tuple, TypeVar
import concurrent.futures
import logging
import time
import hashlib
from collections import OrderedDict
import threading

# Configure logging
logger = logging.getLogger(__name__)

# Initialize parser
parser = Parser()

# Valid borough codes
VALID_BOROUGH_CODES = {1, 2, 3, 4, 5}

# Type variables for better generic typing
T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

# Type definitions for common structures
AddressResult = Dict[str, Any]
AddressList = List[AddressResult]
GeoJSON = Dict[str, Any]
NormalizedAddress = Dict[str, Any]
CoordinatePair = Tuple[float, float]


class ThreadSafeMemoryCache:
    """Thread-safe in-memory LRU cache with TTL."""

    def __init__(self, max_size: int = 1000, ttl_seconds: int = 3600):
        self.max_size = max_size
        self.ttl = ttl_seconds
        self.cache: Dict[str, Tuple[Any, float]] = OrderedDict()
        self._lock = threading.RLock()  # Reentrant lock for thread safety

    def _get_key(self, *args, **kwargs) -> str:
        """Generate a unique key for the function arguments."""
        key_parts = [str(args), str(sorted(kwargs.items()))]
        key_str = "".join(key_parts)
        return hashlib.md5(key_str.encode()).hexdigest()

    def get(self, key: str) -> Optional[Any]:
        """Thread-safe get from cache."""
        with self._lock:
            if key not in self.cache:
                return None

            value, expiry = self.cache[key]

            if time.time() > expiry:
                del self.cache[key]
                return None

            self.cache.move_to_end(key)
            return value

    def set(self, key: str, value: Any) -> None:
        """Thread-safe add to cache."""
        with self._lock:
            expiry = time.time() + self.ttl

            if key in self.cache:
                self.cache[key] = (value, expiry)
                self.cache.move_to_end(key)
                return

            if len(self.cache) >= self.max_size:
                self.cache.popitem(last=False)

            self.cache[key] = (value, expiry)

    def clear(self) -> None:
        """Thread-safe clear cache."""
        with self._lock:
            self.cache.clear()

    def remove_expired(self) -> int:
        """Thread-safe removal of expired items."""
        with self._lock:
            now = time.time()
            expired_keys = [k for k, (_, exp) in self.cache.items() if exp < now]
            for k in expired_keys:
                del self.cache[k]
            return len(expired_keys)


# Function decorator for caching
def cached_method(cache_instance):
    """Decorator to cache method results."""

    def decorator(func):
        def wrapper(self, *args, **kwargs):
            # Only use cache if it's enabled
            if not hasattr(self, "use_cache") or not self.use_cache:
                return func(self, *args, **kwargs)

            cache = getattr(self, cache_instance)
            if cache is None:
                return func(self, *args, **kwargs)

            # Generate cache key
            key = cache._get_key(*args, **kwargs)

            # Try to get from cache
            cached_result = cache.get(key)
            if cached_result is not None:
                return cached_result

            # If not in cache, call the function
            result = func(self, *args, **kwargs)

            # Cache the result
            cache.set(key, result)
            return result

        return wrapper

    return decorator


class AddressFormatter:
    """Consistent formatting for address components."""

    @staticmethod
    def format_borough(borough_name: Optional[str]) -> str:
        """Format borough name consistently."""
        if not borough_name:
            return ""
        return borough_name.upper()

    @staticmethod
    def format_bbl(bbl_data: Optional[Union[Dict[str, Any], str]]) -> Optional[str]:
        """Extract BBL consistently from different formats."""
        if not bbl_data:
            return None

        if isinstance(bbl_data, dict):
            return bbl_data.get("BOROUGH BLOCK LOT (BBL)")
        return str(bbl_data)

    @staticmethod
    def format_coordinates(
        lat: Optional[Union[str, float]], lon: Optional[Union[str, float]]
    ) -> Optional[Dict[str, float]]:
        """Format coordinates consistently."""
        if lat is None or lon is None:
            return None

        try:
            return {"latitude": float(lat), "longitude": float(lon)}
        except (ValueError, TypeError):
            return None


class GeosupportSuggest:
    """Provides address suggestions from NYC Geosupport."""

    def __init__(
        self,
        geosupport=None,
        func="AP",
        max_workers=3,
        rate_limit=0,
        parser_options=None,
        use_cache=False,
        cache_size=1000,
        cache_ttl=3600,
    ):
        """
        Initialize GeosupportSuggest.

        Args:
            geosupport: Geosupport object
            func: Function to use ('AP' or '1B')
            max_workers: Max parallel workers
            rate_limit: Seconds between API calls (0 for no limit)
            parser_options: Dictionary of options to pass to nyc-parser
            use_cache: Enable caching of results
            cache_size: Maximum number of items in memory cache
            cache_ttl: Time-to-live in seconds for cached items
        """
        self._g = geosupport
        self.geofunction = func
        self.results = []
        self.similar_names = []
        self.max_workers = max_workers
        self.rate_limit = rate_limit
        self.last_call_time = 0

        # Initialize parser with custom options if provided
        if parser_options:
            self.parser = Parser(**parser_options)
        else:
            self.parser = parser

        # Initialize cache
        self.use_cache = use_cache
        if use_cache:
            self.cache = ThreadSafeMemoryCache(
                max_size=cache_size, ttl_seconds=cache_ttl
            )
        else:
            self.cache = None

        if self._g is None:
            raise ValueError(
                "You must initialize GeosupportSuggest with a Geosupport object."
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.clear()

    def clear(self):
        """Clear all results and similar names."""
        self.results = []
        self.similar_names = []

    def _respect_rate_limit(self):
        """Implement rate limiting if enabled."""
        if self.rate_limit > 0:
            elapsed = time.time() - self.last_call_time
            if elapsed < self.rate_limit:
                time.sleep(self.rate_limit - elapsed)
            self.last_call_time = time.time()

    def _geocode(self, phn, street, borough_code=None, zip=None):
        """Geocode or attempt to geocode an address."""
        self._respect_rate_limit()
        logger.debug(f"Geocoding: {phn} {street} (Borough: {borough_code}, ZIP: {zip})")

        # Validate borough code
        if borough_code and borough_code not in VALID_BOROUGH_CODES:
            logger.warning(f"Invalid borough code: {borough_code}")
            return

        try:
            r = self._g[self.geofunction](
                house_number=phn, street=street, borough_code=borough_code, zip=zip
            )
            self.results.append(r)
            logger.debug(
                f"Found result: {r.get('First Borough Name', 'Unknown')} - "
                f"{r.get('First Street Name Normalized', 'Unknown')}"
            )
        except GeosupportError as ge:
            if "SIMILAR NAMES" in ge.result.get("Message", ""):
                list_of_street_names = ge.result.get("List of Street Names", [])
                r = [
                    {"street": s, "borough_code": borough_code}
                    for s in list_of_street_names
                ]
                self.similar_names.extend(r)
                logger.debug(f"Found {len(list_of_street_names)} similar names")
            else:
                logger.warning(f"Geocoding error: {ge}")

    def _geocode_parallel(self, items):
        """Geocode multiple items in parallel."""
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.max_workers
        ) as executor:
            futures = []
            for item in items:
                futures.append(
                    executor.submit(
                        self._geocode,
                        item.get("phn"),
                        item.get("street"),
                        item.get("borough_code"),
                        item.get("zip"),
                    )
                )
            concurrent.futures.wait(futures)

    @cached_method("cache")
    def suggestions(
        self,
        input_address: str,
        borough_code: Optional[int] = None,
        parallel: bool = False,
    ) -> AddressList:
        """
        Get valid address suggestions from Geosupport.

        Args:
            input_address: Address string
            borough_code: Borough Code (1-5)
            parallel: Whether to use parallel processing

        Returns:
            List of valid addresses

        Raises:
            ValueError: If borough_code is invalid
        """
        parsed = self.parser.address(input_address)
        if borough_code:
            if borough_code not in VALID_BOROUGH_CODES:
                raise ValueError(
                    f"Invalid borough code: {borough_code}. Must be one of {VALID_BOROUGH_CODES}"
                )
            parsed["BOROUGH_CODE"] = borough_code

        self.similar_names = []
        self.results = []

        if not parsed.get("PHN") or not parsed.get("STREET"):
            logger.warning("No house number or street found in input")
            return self.results

        self._process_address_with_location_info(parsed, parallel)
        self._process_similar_names(parsed, parallel)

        # Filter None values and sort results
        self.results = [r for r in self.results if r is not None]
        self.results.sort(key=lambda x: x.get("First Borough Name", ""))
        return self.results

    def _process_address_with_location_info(self, parsed, parallel):
        """Process address based on available location information."""
        if not parsed.get("BOROUGH_CODE") and not parsed.get("ZIP"):
            self._process_all_boroughs(parsed, parallel)
        elif parsed.get("BOROUGH_CODE"):
            self._geocode(
                phn=parsed["PHN"],
                street=parsed["STREET"],
                borough_code=parsed["BOROUGH_CODE"],
            )
        elif parsed.get("ZIP"):
            self._geocode(phn=parsed["PHN"], street=parsed["STREET"], zip=parsed["ZIP"])

    def _process_all_boroughs(self, parsed, parallel):
        """Try the address in all five boroughs."""
        if parallel:
            items = [
                {
                    "phn": parsed["PHN"],
                    "street": parsed["STREET"],
                    "borough_code": x,
                }
                for x in range(1, 6)
            ]
            self._geocode_parallel(items)
        else:
            for x in range(1, 6):
                self._geocode(
                    phn=parsed["PHN"], street=parsed["STREET"], borough_code=x
                )

    def _process_similar_names(self, parsed, parallel):
        """Process any similar street names returned from Geosupport."""
        if not self.similar_names:
            return

        if parallel:
            items = [
                {
                    "phn": parsed["PHN"],
                    "street": name["street"],
                    "borough_code": name["borough_code"],
                }
                for name in self.similar_names
            ]
            self._geocode_parallel(items)
        else:
            for name in self.similar_names:
                self._geocode(
                    phn=parsed["PHN"],
                    street=name["street"],
                    borough_code=name["borough_code"],
                )

    def suggestions_batch(self, addresses, parallel=False):
        """Process multiple addresses in batch."""
        all_results = []
        for address in addresses:
            if isinstance(address, dict):
                addr_str = address.get("address", "")
                boro = address.get("borough_code")
            else:
                addr_str = address
                boro = None

            results = self.suggestions(addr_str, borough_code=boro, parallel=parallel)
            all_results.append(results)
        return all_results

    def format_address(self, result):
        """Format a result as a standard address string."""
        if not result:
            return ""

        borough = result.get("First Borough Name", "")
        house_num = result.get("House Number - Display Format", "")
        street = result.get("First Street Name Normalized", "")

        return f"{house_num} {street}, {borough}, NY"

    def to_geojson(self, results: AddressList) -> GeoJSON:
        """
        Convert results to GeoJSON format.

        Args:
            results: Address results to convert

        Returns:
            GeoJSON feature collection
        """
        features = []
        for r in results:
            # Skip results without coordinates
            if not all(k in r for k in ["Latitude", "Longitude"]):
                continue

            try:
                lon = float(r.get("Longitude", 0))
                lat = float(r.get("Latitude", 0))

                feature = {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": [lon, lat]},
                    "properties": {
                        "address": self.format_address(r),
                        "borough": r.get("First Borough Name", ""),
                        "block": r.get("BOROUGH BLOCK LOT (BBL)", {}).get(
                            "Tax Block", ""
                        ),
                        "lot": r.get("BOROUGH BLOCK LOT (BBL)", {}).get("Tax Lot", ""),
                    },
                }
                features.append(feature)
            except (ValueError, TypeError):
                continue

        return {"type": "FeatureCollection", "features": features}

    def normalize_results(self, results: AddressList) -> List[NormalizedAddress]:
        """
        Normalize results to a consistent format.

        Handles potential missing fields and standardizes structure.
        """
        normalized = []
        for r in results:
            if not r:
                continue

            norm = {
                "house_number": r.get("House Number - Display Format", ""),
                "street": r.get("First Street Name Normalized", ""),
                "borough": AddressFormatter.format_borough(r.get("First Borough Name")),
                "zip": r.get("ZIP Code", ""),
                "coordinates": AddressFormatter.format_coordinates(
                    r.get("Latitude"), r.get("Longitude")
                ),
                "bbl": AddressFormatter.format_bbl(r.get("BOROUGH BLOCK LOT (BBL)")),
            }

            normalized.append(norm)

        return normalized
